# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

import time
import logging
from datetime import datetime
from typing import Dict, List, Callable, Tuple

from ..core import utils
from .csv_logging import CSVLogger
from ..core.enums.dcb_enums import DCBCommand
from .common_application import CommonApplication
from ..core.enums.pm_enums import PMCommand, ChipID
from ..core.packets.command_packet import CommandPacket
from ..core.packets.common_packets import VersionPacket, StreamPacket
from ..core.enums.common_enums import Application, CommonCommand, Stream
from ..core.packets.pm_packets import MCUVersionPacket, SystemInfoPacket, DateTimePacket
from ..core.packets.pm_packets import BatteryInfoPacket, DCBStatusPacket, ChipIDPacket, PMDCBPacket, \
    PMRegisterReadPacket, PMRegisterWritePacket

logger = logging.getLogger(__name__)


class PMApplication(CommonApplication):
    """
    PM Application class.

    .. code-block:: python3
        :emphasize-lines: 4

        from adi_study_watch import SDK

        sdk = SDK("COM4")
        application = sdk.get_pm_application()
    """

    CHIP_ADPD4K = ChipID.ADPD4K
    CHIP_AD5940 = ChipID.AD5940
    CHIP_AD7156 = ChipID.AD7156
    CHIP_ADXL362 = ChipID.ADXL362
    CHIP_ADP5360 = ChipID.ADP5360
    CHIP_NAND_FLASH = ChipID.NAND_FLASH

    STREAM_BATTERY = Stream.BATTERY

    def __init__(self, packet_manager):
        super().__init__(Application.PM, packet_manager)
        self._args = {}
        self._csv_logger = {}
        self._last_timestamp = {}
        self._callback_function = {}
        self._dcb_size = 57

    def device_configuration_block_status(self) -> Dict:
        """
        Display dcb status of all applications.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.device_configuration_block_status()
            print(x["payload"]["adxl_block"], x["payload"]["adpd4000_block"], x["payload"]["ppg_block"])
            # 0 0 0
        """
        request_packet = CommandPacket(self._destination, DCBCommand.QUERY_STATUS_REQ)
        response_packet = DCBStatusPacket(self._destination, DCBCommand.QUERY_STATUS_RES)
        return self._send_packet(request_packet, response_packet)

    def disable_touch_sensor(self) -> Dict:
        """
        Disables touch sensor.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.disable_touch_sensor()
            print(x["payload"]["status"])
            # PMStatus.OK
        """
        request_packet = CommandPacket(self._destination, PMCommand.DEACTIVATE_TOUCH_SENSOR_REQ)
        response_packet = CommandPacket(self._destination, PMCommand.DEACTIVATE_TOUCH_SENSOR_RES)
        return self._send_packet(request_packet, response_packet)

    def enable_touch_sensor(self) -> Dict:
        """
        Enables touch sensor.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.enable_touch_sensor()
            print(x["payload"]["status"])
            # PMStatus.OK
        """
        request_packet = CommandPacket(self._destination, PMCommand.ACTIVATE_TOUCH_SENSOR_REQ)
        response_packet = CommandPacket(self._destination, PMCommand.ACTIVATE_TOUCH_SENSOR_RES)
        return self._send_packet(request_packet, response_packet)

    def enter_boot_loader_mode(self) -> Dict:
        """
        Sets the device to boot loader mode.

        :return: A empty dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            application.enter_boot_loader_mode()
        """
        request_packet = CommandPacket(self._destination, PMCommand.ENTER_BOOTLOADER_REQ)
        return self._send_packet_no_response(request_packet)

    def get_battery_info(self) -> Dict:
        """
        Returns device current battery information.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.get_battery_info()
            print(x["payload"]["battery_status"], x["payload"]["battery_level"])
            # BatteryStatus.COMPLETE 100
        """
        request_packet = CommandPacket(self._destination, PMCommand.GET_BAT_INFO_REQ)
        response_packet = BatteryInfoPacket(self._destination, PMCommand.GET_BAT_INFO_RES)
        return self._send_packet(request_packet, response_packet)

    def get_chip_id(self, chip_name: ChipID) -> Dict:
        """
        Returns chip id for specified chip name.

        :param chip_name: get chip id of the chip_name, use get_supported_chips() to list all support chip names.
        :type chip_name: ChipID
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.get_supported_chips()
            print(x)
            # [<ChipID.CHIP_ADXL362: ['0x1']>, ... , <ChipID.CHIP_AD7156: ['0x6']>]
            x = application.get_chip_id(application.CHIP_ADPD4K)
            print(x["payload"]["chip_name"], x["payload"]["chip_id"])
            # ChipID.ADPD4K 192
        """
        request_packet = ChipIDPacket(self._destination, PMCommand.CHIP_ID_REQ)
        request_packet.set_payload("chip_name", chip_name)
        response_packet = ChipIDPacket(self._destination, PMCommand.CHIP_ID_RES)
        return self._send_packet(request_packet, response_packet)

    def get_datetime(self) -> Dict:
        """
        Returns device current datetime.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.get_datetime()
            print(f"{x['payload']['year']}-{x['payload']['month']}-{x['payload']['day']}")
            # 2020-12-16
            print(f"{x['payload']['hour']}:{x['payload']['minute']}:{x['payload']['second']}")
            # 15:17:57
        """
        request_packet = CommandPacket(self._destination, PMCommand.GET_DATE_TIME_REQ)
        response_packet = DateTimePacket(self._destination, PMCommand.GET_DATE_TIME_RES)
        return self._send_packet(request_packet, response_packet)

    def get_low_touch_status(self) -> Dict:
        """
        Returns low touch status.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.get_low_touch_status()
            print(x["payload"]["status"])
            # PMStatus.LOW_TOUCH_LOGGING_NOT_STARTED
        """
        request_packet = CommandPacket(self._destination, PMCommand.GET_LOW_TOUCH_LOGGING_STATUS_REQ)
        response_packet = CommandPacket(self._destination, PMCommand.GET_LOW_TOUCH_LOGGING_STATUS_RES)
        return self._send_packet(request_packet, response_packet)

    def get_mcu_version(self) -> Dict:
        """
        Returns Device MCU version.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.get_mcu_version()
            print(x["payload"]["mcu"])
            # MCUType.MCU_M4
        """
        request_packet = CommandPacket(self._destination, PMCommand.GET_MCU_VERSION_REQ)
        response_packet = MCUVersionPacket(self._destination, PMCommand.GET_MCU_VERSION_RES)
        return self._send_packet(request_packet, response_packet)

    @staticmethod
    def get_supported_chips() -> List[ChipID]:
        """
        List all supported chips for PM.

        :return: Array of chips enums.
        :rtype: List

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.get_supported_chips()
            print(x)
            # [<ChipID.CHIP_ADXL362: ['0x1']>, ... , <ChipID.CHIP_AD7156: ['0x6']>]
        """
        return [PMApplication.CHIP_ADXL362, PMApplication.CHIP_ADPD4K, PMApplication.CHIP_ADP5360,
                PMApplication.CHIP_AD5940, PMApplication.CHIP_NAND_FLASH, PMApplication.CHIP_AD7156]

    def get_system_info(self) -> Dict:
        """
        Returns Device system info.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.get_system_info()
            print(x["payload"]["version"])
            # 0
            print(x["payload"]["mac_address"])
            # C5-05-CA-F1-67-D5
            print(x["payload"]["device_id"])
            # 0
        """
        request_packet = CommandPacket(self._destination, PMCommand.SYS_INFO_REQ)
        response_packet = SystemInfoPacket(self._destination, PMCommand.SYS_INFO_RES)
        return self._send_packet(request_packet, response_packet)

    def enable_battery_charging(self) -> Dict:
        """
        Enable battery charging.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            application.enable_battery_charging()
        """
        request_packet = CommandPacket(self._destination, PMCommand.ENABLE_BAT_CHARGE_REQ)
        response_packet = CommandPacket(self._destination, PMCommand.ENABLE_BAT_CHARGE_RES)
        return self._send_packet(request_packet, response_packet)

    def disable_battery_charging(self) -> Dict:
        """
        Disable battery charging.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            application.disable_battery_charging()
        """
        request_packet = CommandPacket(self._destination, PMCommand.DISABLE_BAT_CHARGE_REQ)
        response_packet = CommandPacket(self._destination, PMCommand.DISABLE_BAT_CHARGE_RES)
        return self._send_packet(request_packet, response_packet)

    def get_version(self) -> Dict:
        """
        Returns Device version info.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.get_version()
            print(x["payload"]["major_version"])
            # 1
            print(x["payload"]["minor_version"])
            # 0
            print(x["payload"]["patch_version"])
            # 1
            print(x["payload"]["version_string"])
            # -Perseus
            print(x["payload"]["build_version"])
            # |298b4ce1_Rl|2020-12-14 12:34:31 -0500
        """
        request_packet = CommandPacket(self._destination, CommonCommand.GET_VERSION_REQ)
        response_packet = VersionPacket(self._destination, CommonCommand.GET_VERSION_RES)
        return self._send_packet(request_packet, response_packet)

    def read_register(self, addresses: List[int]) -> Dict:
        """
        Reads the register values of specified addresses. This function takes a list of addresses to read,
        and returns a response packet as dictionary containing addresses and values.

        :param addresses: List of register addresses to read.
        :type addresses: List[int]
        :return: A response packet as dictionary
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x00
             - 0x36

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.read_register([0x15, 0x20, 0x2E])
            print(x["payload"]["data"])
            # [['0x15', '0x0'], ['0x20', '0x0'], ['0x2E', '0x0']]

        """
        data = [[address, 0] for address in addresses]
        request_packet = PMRegisterReadPacket(self._destination, CommonCommand.REGISTER_READ_REQ)
        request_packet.set_payload("size", len(data))
        request_packet.set_payload("data", data)
        response_packet = PMRegisterReadPacket(self._destination, CommonCommand.REGISTER_READ_RES)
        return self._send_packet(request_packet, response_packet)

    def set_datetime(self, date_time: datetime) -> Dict:
        """
        Set specified datetime to device.

        :param date_time: datetime for device.
        :type date_time: datetime
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK
            import datetime

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            now = datetime.datetime.now()
            x = application.set_datetime(now)
            print(x["payload"]["status"])
            # CommonStatus.OK
        """
        is_dst = time.daylight and time.localtime().tm_isdst > 0
        utc_offset = - (time.altzone if is_dst else time.timezone)
        request_packet = DateTimePacket(self._destination, PMCommand.SET_DATE_TIME_REQ)
        request_packet.set_payload("year", date_time.year)
        request_packet.set_payload("month", date_time.month)
        request_packet.set_payload("day", date_time.day)
        request_packet.set_payload("hour", date_time.hour)
        request_packet.set_payload("minute", date_time.minute)
        request_packet.set_payload("second", date_time.second)
        request_packet.set_payload("tz_sec", utc_offset)
        response_packet = CommandPacket(self._destination, PMCommand.SET_DATE_TIME_RES)
        return self._send_packet(request_packet, response_packet)

    def system_hardware_reset(self) -> Dict:
        """
        Reset device hardware.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.system_hardware_reset()
            print(x["payload"]["status"])
            # PMStatus.OK
        """
        request_packet = CommandPacket(self._destination, PMCommand.SYSTEM_HW_RESET_REQ)
        response_packet = CommandPacket(self._destination, PMCommand.SYSTEM_HW_RESET_RES)
        return self._send_packet(request_packet, response_packet)

    def system_reset(self) -> Dict:
        """
        Reset device system.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.system_reset()
            print(x["payload"]["status"])
            # PMStatus.OK
        """
        request_packet = CommandPacket(self._destination, PMCommand.SYSTEM_RESET_REQ)
        response_packet = CommandPacket(self._destination, PMCommand.SYSTEM_RESET_RES)
        return self._send_packet(request_packet, response_packet)

    def write_register(self, addresses_values: List[List[int]]) -> Dict:
        """
        Writes the register values of specified addresses. This function takes a list of addresses and values to write,
        and returns a response packet as dictionary containing addresses and values.

        :param addresses_values: List of register addresses and values to write.
        :type addresses_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 75
           :header-rows: 1

           * - Address ranges
           * - [0x2-0x7], [0xA-0xE], [0x11-0x22], [0x27-0x2E], [0x30-0x33], [0x36]


        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.write_register([[0x20, 0x1], [0x21, 0x2], [0x2E, 0x3]])
            print(x["payload"]["data"])
            # [['0x20', '0x1'], ['0x21', '0x2'], ['0x2E', '0x3']]
        """
        size = len(addresses_values)
        for i in range(size):
            if not (0x2 <= addresses_values[i][0] <= 0x7 or 0xA <= addresses_values[i][0] <= 0xE or
                    0x11 <= addresses_values[i][0] <= 0x22 or 0x27 <= addresses_values[i][0] <= 0x2E or
                    0x30 <= addresses_values[i][0] <= 0x33 or addresses_values[i][0] == 0x36):
                logger.warning(f"{'0x%X' % addresses_values[i][0]} is out of range, allowed ranges are: [0x2-0x7], "
                               f"[0xA-0xE], [0x11-0x22], [0x27-0x2E], [0x30-0x33], [0x36]")
        request_packet = PMRegisterWritePacket(self._destination, CommonCommand.REGISTER_WRITE_REQ)
        request_packet.set_payload("size", size)
        request_packet.set_payload("data", addresses_values)
        response_packet = PMRegisterWritePacket(self._destination, CommonCommand.REGISTER_WRITE_RES)
        return self._send_packet(request_packet, response_packet)

    def delete_device_configuration_block(self) -> Dict:
        """
        Deletes PM Device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            application.delete_device_configuration_block()
        """
        request_packet = CommandPacket(self._destination, DCBCommand.ERASE_CONFIG_REQ)
        response_packet = PMDCBPacket(self._destination, DCBCommand.ERASE_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def read_device_configuration_block(self) -> Dict:
        """
        Returns entire device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.read_device_configuration_block()
            print(x["payload"]["dcb_data"])
            # []
        """
        request_packet = CommandPacket(self._destination, DCBCommand.READ_CONFIG_REQ)
        response_packet = PMDCBPacket(self._destination, DCBCommand.READ_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def write_device_configuration_block(self, addresses_values: List[List[int]]) -> Dict:
        """
        Writes the device configuration block values of specified addresses.
        This function takes a list of addresses and values to write, and returns a response packet as
        dictionary containing addresses and values.

        :param addresses_values: List of addresses and values to write.
        :type addresses_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.write_device_configuration_block([[0x2, 2], [0x1, 0x1]])
            print(x["payload"]["size"])
            # 2
        """
        request_packet = PMDCBPacket(self._destination, DCBCommand.WRITE_CONFIG_REQ)
        request_packet.set_payload("size", len(addresses_values))
        request_packet.set_payload("data", addresses_values)
        response_packet = PMDCBPacket(self._destination, DCBCommand.WRITE_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def write_device_configuration_block_from_file(self, filename: str) -> Dict:
        """
        Writes the device configuration block values of specified addresses from file.

        :param filename: dcb filename
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            application.write_device_configuration_block_from_file("pm_dcb.dcfg")
        """
        result = self.device_configuration_file_to_list(filename)
        if result:
            return self.write_device_configuration_block(result)

    # noinspection PyTypeChecker
    def set_callback(self, callback_function: Callable, args: Tuple = (), stream: Stream = STREAM_BATTERY) -> None:
        """
        Sets the callback for the stream data.

        :param callback_function: callback function for specified adpd stream.
        :param args: optional arguments that will be passed with the callback.
        :param stream: Callback for specified stream.
        :type stream: Stream
        :return: None

        .. code-block:: python3
            :emphasize-lines: 4,12

            from adi_study_watch import SDK

            # make sure optional arguments have default value to prevent them causing Exceptions.
            def callback(data, optional1=None, optional2=None):
                print(data)

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            # these optional arguments can be used to pass file, matplotlib or other objects to manipulate data.
            optional_arg1 = "1"
            optional_arg2 = "2"
            application.set_callback(callback, args=(optional_arg1, optional_arg2))
        """
        self._callback_function[stream] = callback_function
        self._args[stream] = args

    def _callback_data(self, packet, packet_id, callback_function=None, args=None, stream=None):
        """
        Process and returns the data back to user's callback function.
        """
        self._callback_data_helper(packet, BatteryInfoPacket(), self.STREAM_BATTERY)

    def _callback_data_helper(self, packet, response_packet, stream=None):
        """
        Process and returns the data back to user's callback function.
        """
        args = self._args.get(stream, ())
        callback_function = self._callback_function.get(stream, None)
        response_packet.decode_packet(packet)
        last_timestamp = self._last_timestamp.get(stream)
        result = response_packet.get_dict(last_timestamp)
        utils.update_timestamp(result, last_timestamp)
        csv_logger = self._csv_logger.get(stream, None)
        if csv_logger:
            csv_logger.add_row(result, last_timestamp[0])
        try:
            if callback_function:
                callback_function(result, *args)
        except Exception as e:
            logger.error(f"Can't send packet back to user callback function, reason :: {e}", exc_info=True)

        if not csv_logger and not callback_function:
            logger.warning(f"No callback function provided for {result['header']['source']}")

    def subscribe_stream(self, stream=STREAM_BATTERY):
        """
        Subscribe to the stream.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            subs_stream = application.subscribe_stream()
            print(subs_stream["payload"]["status"])
            # CommonStatus.SUBSCRIBER_ADDED
        """
        request_packet = StreamPacket(self._destination, CommonCommand.SUBSCRIBE_STREAM_REQ)
        request_packet.set_payload("stream_address", stream)
        self._subscribe_stream_data(stream)
        response_packet = StreamPacket(self._destination, CommonCommand.SUBSCRIBE_STREAM_RES)
        self._update_timestamp(datetime.now(), stream)
        return self._send_packet(request_packet, response_packet)

    def _update_timestamp(self, date_time, stream=STREAM_BATTERY, generate_ts=False):
        if generate_ts:
            ts = (32000.0 * ((date_time.hour * 3600) + (date_time.minute * 60) + date_time.second))
        else:
            ts = -1
        self._last_timestamp[stream] = [date_time.timestamp(), ts]

    def unsubscribe_stream(self, stream=None):
        """
        Unsubscribe the stream.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            unsubscribe_stream = application.unsubscribe_stream()
            print(unsubscribe_stream["payload"]["status"])
            # CommonStatus.SUBSCRIBER_REMOVED
        """
        request_packet = StreamPacket(self._destination, CommonCommand.UNSUBSCRIBE_STREAM_REQ)
        request_packet.set_payload("stream_address", stream)
        response_packet = StreamPacket(self._destination, CommonCommand.UNSUBSCRIBE_STREAM_RES)
        response_packet = self._send_packet(request_packet, response_packet)
        self._unsubscribe_stream_data(stream)
        return response_packet

    def _subscribe_stream_data(self, stream=None, callback_function=None):
        callback_function = callback_function if callback_function else self._callback_data
        data_packet_id = self._get_packet_id(CommonCommand.STREAM_DATA, stream)
        self._packet_manager.subscribe(data_packet_id, callback_function)

    def _unsubscribe_stream_data(self, stream=None, callback_function=None):
        callback_function = callback_function if callback_function else self._callback_data
        data_packet_id = self._get_packet_id(CommonCommand.STREAM_DATA, stream)
        self._packet_manager.unsubscribe(data_packet_id, callback_function)

    def enable_csv_logging(self, filename, header=None) -> None:
        """
        Start logging stream data into CSV.

        :param filename: Name of the CSV file.
        :param header: Header list of the CSV file.
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.enable_csv_logging("adp.csv")
        """
        if header is None:
            header = ["Timestamp", "Battery status", "Battery level", "Battery mv"]
        self._csv_logger[Stream.BATTERY] = CSVLogger(filename, header)

    def disable_csv_logging(self) -> None:
        """
        Stops logging stream data into CSV.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.disable_csv_logging()
        """
        if self._csv_logger.get(Stream.BATTERY):
            self._csv_logger[Stream.BATTERY].stop_logging()
        self._csv_logger[Stream.BATTERY] = None



from ..data_types.array import Array
from ..data_types.integer import Int
from .command_packet import CommandPacket


class AD7156RegisterReadPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.AD7156: ['0xC8', '0x0B']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 23,
                'checksum': 0
            },
            'payload': {
                'command': <CommonCommand.REGISTER_READ_RES: ['0x22']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'size': 3,
                'data': [
                    [ '0x10', '0x40' ],
                    [ '0x11', '0xC0' ],
                    [ '0x12', '0xC0' ]
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["size"] = Int(1)
        self._config["payload"]["data"] = Array(-1, dimension=2,
                                                data_types=[
                                                    Int(2, value_limit=[0x00, 0x17], to_hex=True),
                                                    Int(2, to_hex=True)
                                                ])


class AD7156RegisterWritePacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.AD7156: ['0xC8', '0x0B']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 15,
                'checksum': 0
            },
            'payload': {
                'command': <CommonCommand.REGISTER_WRITE_RES: ['0x24']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'size': 1,
                'data': [
                    [ '0x10', '0x11' ]
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["size"] = Int(1)
        self._config["payload"]["data"] = Array(-1, dimension=2,
                                                data_types=[
                                                    Int(2, value_limit=[0x09, 0x12], to_hex=True),
                                                    Int(2, to_hex=True)
                                                ])


class AD7156DCBPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.AD7156: ['0xC8', '0x0B']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 92,
                'checksum': 0
            },
            'payload': {
                'command': <DCBCommand.READ_CONFIG_RES: ['0x98']>,
                'status': <DCBStatus.OK: ['0x97']>,
                'size': 7,
                'data': [
                    [ '0x9', '0x19' ],
                    [ '0xA', '0x66' ],
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["size"] = Int(2)
        self._config["payload"]["data"] = Array(-1, dimension=2,
                                                data_types=[
                                                    Int(1, to_hex=True),
                                                    Int(3, value_limit=[0x09, 0x12], to_hex=True)
                                                ], reverse_inner_array=True)



from enum import Enum, unique

from .. import utils


@unique
class User0Command(Enum):
    """
    User0 Command Enum
    """
    LOWEST = [0x40]
    SET_STATE_REQ = [0x42]
    SET_STATE_RES = [0x43]
    GET_STATE_REQ = [0x44]
    GET_STATE_RES = [0x45]
    ID_OP_REQ = [0x46]
    ID_OP_RES = [0x47]
    CLEAR_PREV_ST_EVT_REQ = [0x48]
    CLEAR_PREV_ST_EVT_RES = [0x49]
    GET_PREV_ST_EVT_REQ = [0x4A]
    GET_PREV_ST_EVT_RES = [0x4B]
    BYPASS_USER0_TIMINGS_REQ = [0x4C]
    BYPASS_USER0_TIMINGS_RES = [0x4D]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


@unique
class User0Status(Enum):
    """
    User0 status Enum
    """
    LOWEST = [0x40]
    OK = [0x41]
    ERR_ARGS = [0x42]
    ERR_NOT_CHKD = [0xFF]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


@unique
class User0State(Enum):
    """
    User0 status Enum
    """
    ADMIT_STANDBY = [0x0]
    START_MONITORING = [0x1]
    SLEEP = [0x2]
    INTERMITTENT_MONITORING = [0x3]
    INTERMITTENT_MONITORING_START_LOG = [0x4]
    INTERMITTENT_MONITORING_STOP_LOG = [0x5]
    END_MONITORING = [0x6]
    CHARGING_BATTERY = [0x7]
    OUT_OF_BATTERY_STATE_BEFORE_START_MONITORING = [0x8]
    OUT_OF_BATTERY_STATE_DURING_INTERMITTENT_MONITORING = [0x9]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


@unique
class User0Event(Enum):
    """
    User0 event Enum
    """
    INVALID = [0x0]
    NAV_BUTTON_RESET = [0x1]
    WATCH_ON_CRADLE_NAV_BUTTON_RESET = [0x2]
    BATTERY_DRAINED = [0x3]
    BLE_DISCONNECT_UNEXPECTED = [0x4]
    BLE_DISCONNECT_NW_TERMINATED = [0x5]
    WATCH_OFF_CRADLE_BLE_DISCONNECT_NW_TERMINATED = [0x6]
    RTC_TIMER_INTERRUPT = [0x7]
    BLE_ADV_TIMEOUT = [0x8]
    USB_DISCONNECT_UNEXPECTED = [0x9]
    BATTERY_FULL = [0xA]
    FINISH_LOG_TRANSFER = [0xB]
    WATCH_OFF_CRADLE_BLE_CONNECTION = [0xC]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


@unique
class User0ID(Enum):
    """
    User0 ID Enum
    """
    HW_ID = [0x0]
    EXP_ID = [0x1]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


@unique
class User0OperationMode(Enum):
    """
    User0 operation Enum
    """
    READ = [0x0]
    WRITE = [0x1]
    DELETE = [0x2]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from ..data_types.integer import Int
from ..data_types.enums import Enums
from ..data_types.string import String
from ..enums.common_enums import Stream
from .command_packet import CommandPacket


class DecimationFactorPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xD',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.GET_STREAM_DEC_FACTOR_RES: ['0x2A']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'stream_address': <Stream.ADPD6: ['0xC2', '0x16']>,
                'decimation_factor': 1
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["stream_address"] = Enums(2, Stream, reverse=True)
        self._config["payload"]["decimation_factor"] = Int(1, value_limit=[1, 5])


class StreamPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADXL: ['0xC1', '0x02']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xC',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.SUBSCRIBE_STREAM_RES: ['0x0D']>,
                'status': <CommonStatus.SUBSCRIBER_ADDED: ['0x09']>,
                'stream_address': <Stream.ADXL: ['0xC2', '0x02']>
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["stream_address"] = Enums(2, Stream, reverse=True, default=Stream.NULL)


class StreamStatusPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADXL: ['0xC1', '0x02']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xE',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.GET_SENSOR_STATUS_RES: ['0x11']>,
                'status': <CommonStatus.STREAM_STOPPED: ['0x03']>,
                'stream_address': <Stream.ADXL: ['0xC2', '0x02']>,
                'num_subscribers': 0,
                'num_start_registered': 0
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["stream_address"] = Enums(2, Stream, reverse=True)
        self._config["payload"]["num_subscribers"] = Int(1)
        self._config["payload"]["num_start_registered"] = Int(1)


class VersionPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.ADPD: ['0xC1', '0x10']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x42',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.GET_VERSION_RES: ['0x01']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'major_version': 0,
                'minor_version': 3,
                'patch_version': 1,
                'version_string': 'ADPD_App',
                'build_version': 'TEST ADPD4000_VERSION STRING'
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["major_version"] = Int(2)
        self._config["payload"]["minor_version"] = Int(2)
        self._config["payload"]["patch_version"] = Int(2)
        self._config["payload"]["version_string"] = String(10)
        self._config["payload"]["build_version"] = String(40)



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from enum import Enum, unique

from .. import utils


@unique
class FSCommand(Enum):
    """
    FSCommand Enum
    """
    LOWEST = [0x40]
    MOUNT_REQ = [0x42]
    MOUNT_RES = [0x43]
    FORMAT_REQ = [0x46]
    FORMAT_RES = [0x47]
    LS_REQ = [0x48]
    LS_RES = [0x49]
    VOL_INFO_REQ = [0x4E]
    VOL_INFO_RES = [0x4F]
    GET_STATUS_REQ = [0x50]
    GET_STATUS_RES = [0x51]
    GET_STREAM_SUB_STATUS_REQ = [0x52]
    GET_STREAM_SUB_STATUS_RES = [0x53]
    STREAM_DEBUG_INFO_REQ = [0x56]
    STREAM_DEBUG_INFO_RES = [0x57]
    TEST_LOG_REQ = [0x58]
    TEST_LOG_RES = [0x59]
    FORCE_STOP_LOG_REQ = [0x5A]
    FORCE_STOP_LOG_RES = [0x5B]
    SET_KEY_VALUE_PAIR_REQ = [0x64]
    SET_KEY_VALUE_PAIR_RES = [0x65]
    DCFG_START_LOG_REQ = [0x66]
    DCFG_START_LOG_RES = [0x67]
    DCFG_STOP_LOG_REQ = [0x68]
    DCFG_STOP_LOG_RES = [0x69]
    LOG_USER_CONFIG_DATA_REQ = [0x6A]
    LOG_USER_CONFIG_DATA_RES = [0x6B]
    DELETE_CONFIG_FILE_REQ = [0x6E]
    DELETE_CONFIG_FILE_RES = [0x6F]
    GET_NUMBER_OF_FILE_REQ = [0x70]
    GET_NUMBER_OF_FILE_RES = [0x71]
    START_STREAM_LOGGING_REQ = [0x72]
    START_STREAM_LOGGING_RES = [0x73]
    STOP_STREAM_LOGGING_REQ = [0x74]
    STOP_STREAM_LOGGING_RES = [0x75]
    START_LOGGING_REQ = [0x76]
    START_LOGGING_RES = [0x77]
    STOP_LOGGING_REQ = [0x78]
    STOP_LOGGING_RES = [0x79]
    DOWNLOAD_LOG_REQ = [0x7A]
    DOWNLOAD_LOG_RES = [0x7B]
    GET_BAD_BLOCKS_REQ = [0x7E]
    GET_BAD_BLOCKS_RES = [0x7F]
    CHUNK_RETRANSMIT_REQ = [0x84]
    CHUNK_RETRANSMIT_RES = [0x85]
    GET_DEBUG_INFO_REQ = [0x88]
    GET_DEBUG_INFO_RES = [0x89]
    PATTERN_WRITE_REQ = [0x8A]
    PATTERN_WRITE_RES = [0x8B]
    GET_FILE_INFO_REQ = [0x8C]
    GET_FILE_INFO_RES = [0x8D]
    PAGE_READ_TEST_REQ = [0x8E]
    PAGE_READ_TEST_RES = [0x8F]
    APPEND_FILE_REQ = [0xA0]
    APPEND_FILE_RES = [0xA1]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


@unique
class FSStatus(Enum):
    """
    FSStatus Enum
    """
    LOWEST = [0x40]
    OK = [0x41]
    ERROR = [0x42]
    END_OF_FILE = [0x43]
    END_OF_DIR = [0x44]
    ERR_INVALID = [0x45]
    ERR_ARGS = [0x46]
    ERR_FORMAT = [0x47]
    ERR_MEMORY_FULL = [0x48]
    ERR_LOG_FORCE_STOPPED = [0x49]
    ERR_MAX_FILE_COUNT = [0x4A]
    CONFIG_FILE_FOUND = [0x4B]
    CONFIG_FILE_NOT_FOUND = [0x4C]
    LOGGING_STOPPED = [0x4D]
    LOGGING_IN_PROGRESS = [0x4E]
    LOGGING_ERROR = [0x4F]
    LOGGING_NOT_STARTED = [0x50]
    ERR_BATTERY_LOW = [0x51]
    ERR_POWER_STATE_SHUTDOWN = [0x52]
    ERR_CONFIG_FILE_POSITION = [0x53]
    BLOCKS_WRITE_ERROR = [0x54]
    NO_FILE_TO_APPEND = [0x55]
    ERR_NOT_CHKD = [0xFF]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


@unique
class FileType(Enum):
    """
    FileType Enum
    """
    CONFIG_FILE = [0x0]
    DATA_FILE = [0x1]
    INVALID_TYPE = [0x2]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


@unique
class FSLogging(Enum):
    """
    FSLogging Enum
    """
    STOP_LOGGING = [0x0]
    MEMORY_FULL = [0x1]
    BATTERY_LOW = [0x2]
    POWER_STATE_SHUTDOWN = [0x3]
    STREAMS_NOT_UNSUBSCRIBED = [0x16]
    STOP_LOGGING_INVALID = [0xFF]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


@unique
class FSSubState(Enum):
    """
    FSSubState Enum
    """
    UNSUBSCRIBED = [0x0]
    SUBSCRIBED = [0x1]
    INVALID = [0xFF]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))



import struct
import logging
from typing import List
from .data_type import DataType

logger = logging.getLogger(__name__)


class Decimal(DataType):

    def __init__(self, size, reverse: bool = False, value_range: List = None, default: float = 0.0):
        super().__init__(size, reverse, default)
        self._value_range = value_range

    def encode(self):
        result = list(struct.pack("f", self.get_value()))
        return reversed(result) if self._reverse else result

    def decode(self, array: List):
        array = reversed(array) if self._reverse else array
        self.set_value(struct.unpack("f", bytes(array))[0])

    def default_value(self):
        return self.encode(self._default)

    def _value_check(self, value: float):
        if not type(value) == float:
            logger.error(f"{value} is not of type float. Using default value {self._default}.")
            return False

        if self._value_range:
            if not self._value_range[0] <= value <= self._value_range[1]:
                logger.error(f"{'0x%X' % value} is out of range, value needs to be greater than or equal to "
                             f"{'0x%X' % self._value_range[0]} and less than or equal to "
                             f"{'0x%X' % self._value_range[1]}.")
                return False

        return True



import logging
from typing import List

from .data_type import DataType
from .. import utils

logger = logging.getLogger(__name__)


class Int(DataType):

    def __init__(self, size: int, sign: bool = False, reverse: bool = False, value_limit: List = None,
                 default: int = None, to_hex: bool = False):
        super().__init__(size, reverse, default)
        self._sign = sign
        self._to_hex = to_hex
        self._value_limit = value_limit

    def encode(self):
        self._value_check()
        return utils.split_int_in_bytes(self.get_value(), length=self._size, reverse=self._reverse)

    def decode(self, array: List):
        value = utils.join_multi_length_packets(array, sign=self._sign, reverse=self._reverse,
                                                convert_to_hex=self._to_hex)
        self.set_value(value)

    def _default_value(self):
        self.set_value(0)

    def _value_check(self):
        value = self.get_value()

        if value is None:
            self._default_value()
            return

        if not type(value) == int:
            self._default_value()
            logger.error(f"{value} is not of type int. Using default value {self.get_value()}.")
            return False

        if self._value_limit:
            if not self._value_limit[0] <= value <= self._value_limit[1]:
                logger.error(f"{'0x%X' % value} is out of range, value needs to be greater than or equal to "
                             f"{'0x%X' % self._value_limit[0]} and less than or equal to "
                             f"{'0x%X' % self._value_limit[1]}.")
                return False

        lower_bound = 0
        upper_bound = 16 ** (self._size * 2) - 1
        if self._sign:
            lower_bound = -(16 ** ((self._size - 1) * 2))
            upper_bound = 16 ** ((self._size - 1) * 2) - 1

        if not lower_bound <= value <= upper_bound:
            logger.error(f"{'0x%X' % value} is out of range, value needs to be greater than or equal to "
                         f"{'0x%X' % lower_bound} and less than or equal to {'0x%X' % upper_bound}.")
            return False

        return True



import logging
from typing import List
from .data_type import DataType

logger = logging.getLogger(__name__)


class Array(DataType):

    def __init__(self, size: int, reverse_inner_array=None, dimension: int = 1, data_types: List = None,
                 size_limit: int = None, default=None):
        super().__init__(size, reverse_inner_array, default)
        self._dimension = dimension
        self._data_types = data_types
        self._size_limit = size_limit

    def encode(self):
        result = []
        value = self.get_value()
        if value is None:
            return result
        self.reverse_pair_check(value)
        for val in value:
            for i in range(self._dimension):
                temp_val = val if self._dimension == 1 else val[i]
                data_type = self._data_types[i]
                data_type.set_value(temp_val)
                result += data_type.encode()
        return result

    def decode(self, array: List):
        result = []
        total_size = 0
        while not len(array) == 0:
            inner_result = []
            for i in range(self._dimension):
                data_type = self._data_types[i]
                data_size = data_type.get_size()
                total_size += data_size
                value = array[:data_size]
                array = array[data_size:]
                data_type.decode(value)
                inner_result.append(data_type.get_value())
            result.append(inner_result)
        if self._dimension == 1:
            result = [res[0] for res in result]
        self.reverse_pair_check(result)
        self._size = total_size
        self.set_value(result)

    def reverse_pair_check(self, result):
        if self._dimension > 1 and self._reverse:
            for res in result:
                res.reverse()

    def default_value(self):
        return self.encode(self._default)

    def _value_check(self, value: List):
        if not type(value) == list:
            logger.error(f"{value} is not of type list. Using default value {self._default}.")
            return False

        if self._size_limit:
            if len(value) > self._size_limit:
                logger.error(f"Max length of {value} can be {self._size_limit}.")
                return False
        return True



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************
import re
import math
import logging
from typing import List, Dict

from ..core import utils
from .csv_logging import CSVLogger
from .common_stream import CommonStream
from ..core.enums.dcb_enums import DCBCommand
from ..core.packets.command_packet import CommandPacket
from ..core.packets.common_packets import StreamStatusPacket
from ..core.packets.stream_data_packets import TemperatureDataPacket
from ..core.enums.common_enums import Application, Stream, CommonCommand
from ..core.packets.temperature_packets import TemperatureDCFGPacket, TemperatureDCBPacket, \
    TemperatureLibraryConfigPacket

logger = logging.getLogger(__name__)


class TemperatureApplication(CommonStream):
    """
    Temperature Application class.

    .. code-block:: python3
        :emphasize-lines: 4

        from adi_study_watch import SDK

        sdk = SDK("COM4")
        application = sdk.get_temperature_application()

    """

    STREAM_TEMPERATURE1 = Stream.TEMPERATURE1
    STREAM_TEMPERATURE2 = Stream.TEMPERATURE2
    STREAM_TEMPERATURE3 = Stream.TEMPERATURE3
    STREAM_TEMPERATURE4 = Stream.TEMPERATURE4
    STREAM_TEMPERATURE5 = Stream.TEMPERATURE5
    STREAM_TEMPERATURE6 = Stream.TEMPERATURE6
    STREAM_TEMPERATURE7 = Stream.TEMPERATURE7
    STREAM_TEMPERATURE8 = Stream.TEMPERATURE8
    STREAM_TEMPERATURE9 = Stream.TEMPERATURE9
    STREAM_TEMPERATURE10 = Stream.TEMPERATURE10
    STREAM_TEMPERATURE11 = Stream.TEMPERATURE11
    STREAM_TEMPERATURE12 = Stream.TEMPERATURE12

    def __init__(self, packet_manager):
        super().__init__(Application.TEMPERATURE, Stream.TEMPERATURE, packet_manager)
        self._slope = 1
        self._correction_factor = 0
        self._dcb_size = 57

    def get_device_configuration(self) -> List[Dict]:
        """
        Returns device configuration data.

        :return: A response packet as dictionary.
        :rtype: List[Dict]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            x = application.get_device_configuration()
            print(x[0]["payload"]["data"])
            # [['0x9', '0x97'], ['0x7', '0x8FFF'], ['0xB', '0x2F6'], ... ]
        """
        request_packet = CommandPacket(self._destination, CommonCommand.GET_DCFG_REQ)
        response_packet = TemperatureDCFGPacket(self._destination, CommonCommand.GET_DCFG_RES)
        return self._send_packet_multi_response(request_packet, response_packet)

    def _callback_data(self, packet, packet_id, callback_function=None, args=None):
        """
        Process and returns the data back to user's callback function.
        """
        self._callback_data_helper(packet, TemperatureDataPacket())

    def _update_stream_data(self, result):
        compensated_temperature = (result["payload"]["skin_temperature"] * self._slope) + self._correction_factor
        result["payload"]["compensated_temperature"] = compensated_temperature

    @staticmethod
    def get_supported_streams() -> List[Stream]:
        """
        List all supported streams for Temperature.

        :return: Array of stream ID enums.
        :rtype: List[Stream]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            x = application.get_supported_streams()
            print(x)
            # [<Stream.TEMPERATURE1: ['0xC8', '0x16']>, ... , <Stream.TEMPERATURE12: ['0xC8', '0x21']>]
        """
        return [TemperatureApplication.STREAM_TEMPERATURE1, TemperatureApplication.STREAM_TEMPERATURE2,
                TemperatureApplication.STREAM_TEMPERATURE3, TemperatureApplication.STREAM_TEMPERATURE4,
                TemperatureApplication.STREAM_TEMPERATURE5, TemperatureApplication.STREAM_TEMPERATURE6,
                TemperatureApplication.STREAM_TEMPERATURE7, TemperatureApplication.STREAM_TEMPERATURE8,
                TemperatureApplication.STREAM_TEMPERATURE9, TemperatureApplication.STREAM_TEMPERATURE10,
                TemperatureApplication.STREAM_TEMPERATURE11, TemperatureApplication.STREAM_TEMPERATURE12]

    def _temperature_stream_helper(self, stream: Stream) -> Stream:
        """
        Confirms stream is from list of Enums.
        """
        if stream in self.get_supported_streams():
            return stream
        else:
            logger.warning(f"{stream} is not supported stream, choosing {self.get_supported_streams()[0]} "
                           f"as default stream. use get_supported_streams() to know all supported streams.")
            return self.get_supported_streams()[0]

    def get_sensor_status(self, stream=STREAM_TEMPERATURE1) -> Dict:
        """
        Returns packet with number of subscribers and number of sensor start request registered.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            x = application.get_sensor_status(application.STREAM_TEMPERATURE1)
            print(x["payload"]["num_subscribers"], x["payload"]["num_start_registered"])
            # 0 0
        """
        stream = self._temperature_stream_helper(stream)
        request_packet = StreamStatusPacket(self._destination, CommonCommand.GET_SENSOR_STATUS_REQ)
        request_packet.set_payload("stream_address", stream)
        response_packet = StreamStatusPacket(self._destination, CommonCommand.GET_SENSOR_STATUS_RES)
        return self._send_packet(request_packet, response_packet)

    def write_dcb_to_lcfg(self) -> Dict:
        """
        Writes Device configuration block data to library configuration.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            x = application.write_dcb_to_lcfg()
            print(x["payload"]["status"])
            # CommonStatus.OK
        """
        request_packet = CommandPacket(self._destination, CommonCommand.SET_LCFG_REQ)
        response_packet = CommandPacket(self._destination, CommonCommand.SET_LCFG_RES)
        return self._send_packet(request_packet, response_packet)

    def delete_device_configuration_block(self) -> Dict:
        """
        Deletes PPG Device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            application.delete_device_configuration_block()
        """
        request_packet = CommandPacket(self._destination, DCBCommand.ERASE_CONFIG_REQ)
        response_packet = TemperatureDCBPacket(self._destination, DCBCommand.ERASE_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def read_device_configuration_block(self) -> Dict:
        """
        Returns entire device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            x = application.read_device_configuration_block()
            print(x["payload"]["data"])
            # []

        """
        request_packet = CommandPacket(self._destination, DCBCommand.READ_CONFIG_REQ)
        response_packet = TemperatureDCBPacket(self._destination, DCBCommand.READ_CONFIG_RES)
        response_dict = self._send_packet_multi_response(request_packet, response_packet)
        for packet in response_dict:
            packet["payload"]["data"] = utils.add_index_to_array(packet["payload"]["data"], to_hex=True)
        return response_dict

    def read_library_configuration(self, field: int) -> Dict:
        """
        Reads library configuration from specified field values.

        :param field: field values to read.
        :type field: int
        :return: A response packet as dictionary
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Fields Lower Limit
             - Fields Upper Limit
           * - 0x00
             - 0x14

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            x = application.read_library_configuration([0x00])
            print(x["payload"]["data"])
            # [['0x0', '0x0']]

        """
        request_packet = TemperatureLibraryConfigPacket(self._destination, CommonCommand.READ_LCFG_REQ)
        request_packet.set_payload("field", field)
        response_packet = TemperatureLibraryConfigPacket(self._destination, CommonCommand.READ_LCFG_RES)
        response_dict = self._send_packet(request_packet, response_packet)
        if field in [0, 1]:
            response_dict["payload"]["data"] = response_dict["payload"]["data"][:1]
        return response_dict

    def write_device_configuration_block(self, addresses_values: List[List[int]]) -> List:
        """
        Writes the device configuration block values of specified addresses.
        This function takes a list of addresses and values to write, and returns a response packet as
        dictionary containing addresses and values.

        :param addresses_values: List of addresses and values to write.
        :type addresses_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            x = application.write_device_configuration_block([[0x20, 2], [0x21, 0x1]])
            print(x["payload"]["size"])
            # 2

        """
        dcb_array = []
        for address_value in addresses_values:
            dcb_array.append(address_value[1])
        result = []
        packets = math.ceil(len(dcb_array) / self._dcb_size)
        for packet in range(packets):
            addresses_value = dcb_array[packet * self._dcb_size: (packet + 1) * self._dcb_size]
            request_packet = TemperatureDCBPacket(self._destination, DCBCommand.WRITE_CONFIG_REQ)
            request_packet.set_payload("size", len(addresses_value))
            request_packet.set_payload("packet_count", packets)
            request_packet.set_payload("data", addresses_value)
            response_packet = TemperatureDCBPacket(self._destination, DCBCommand.WRITE_CONFIG_RES)
            result.append(self._send_packet(request_packet, response_packet))
        return result

    def write_device_configuration_block_from_file(self, filename: str) -> List:
        """
        Writes the device configuration block values of specified addresses from file.

        :param filename: dcb filename
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            application.write_device_configuration_block_from_file("temperature_dcb.lcfg")

        """
        file = open(filename, 'r')
        regex = r"^<.*?>(?:.|\n)+?<\/.*?>"
        test_str = "".join(file.readlines())
        file.close()
        matches = re.findall(regex, test_str, re.MULTILINE)
        dcb = {}
        for x in matches:
            lines = x.split("\n")
            first_line = lines[0].strip()
            dcb[first_line] = lines[1:-1]

        if dcb.get("<TEMP_DCB>", None) is None:
            raise Exception("Invalid file format.")

        result = []
        selected_dcb = 0
        for line in dcb["<TEMP_DCB>"]:
            dcb_value = self._parse_single_dcb_line(line)
            result.append([int(dcb_value[0], 16), int(dcb_value[1], 16)])
            if int(dcb_value[0], 16) == 1:
                selected_dcb = format(int(dcb_value[1], 16), 'b').zfill(16)
        selected_dcb = list(reversed(selected_dcb))

        required_dcb = []
        for i, value in enumerate(selected_dcb):
            slot = (i + 65) * int(value)
            if slot and chr(slot) in ['C', 'D', 'J', 'K', 'L']:
                dcb_slot = f"<TEMP_SLOT{chr(slot)}_DCB>"
                required_dcb.append(dcb_slot)
                if dcb.get(dcb_slot, None) is None:
                    raise Exception(f"{dcb_slot} not present.")

        for dcb_slot in required_dcb:
            dcb_lines = dcb[dcb_slot]
            for line in dcb_lines:
                if line[0] != '#' and line[0] != '\n' and line[0] != ' ' and line[0] != '\t':
                    dcb_value = self._parse_single_dcb_line(line)
                    result.append([int(dcb_value[0], 16), int(dcb_value[1], 16)])
        if result:
            return self.write_device_configuration_block(result)

    def write_library_configuration(self, field, value) -> Dict:
        """
        Writes library configuration from List of fields and values.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Fields Lower Limit
             - Fields Upper Limit
           * - 0x00
             - 0x14

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            x = application.write_library_configuration([[0x00, 0x1]])
            print(x["payload"]["data"])
            # [['0x0', '0x1']]

        """
        if field in [0, 1]:
            if not type(value) is int:
                raise Exception("Value needs to be int for fields 0 and 1.")
            data = [value]
        else:
            if not type(value) is str:
                raise Exception("Value needs to be file address (str) for fields other than 0 and 1.")
            addresses_values = self.device_configuration_file_to_list(value)
            data = []
            for address_value in addresses_values:
                data.append(address_value[1])

        request_packet = TemperatureLibraryConfigPacket(self._destination, CommonCommand.WRITE_LCFG_REQ)
        request_packet.set_payload("field", field)
        request_packet.set_payload("data", data)
        response_packet = TemperatureLibraryConfigPacket(self._destination, CommonCommand.WRITE_LCFG_RES)
        response_dict = self._send_packet(request_packet, response_packet)
        if field in [0, 1]:
            response_dict["payload"]["data"] = response_dict["payload"]["data"][:1]
        return response_dict

    def enable_csv_logging(self, filename, header=None) -> None:
        """
        Start logging stream data into CSV.

        :param filename: Name of the CSV file.
        :param header: Header list of the CSV file.
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            x = application.enable_csv_logging("temp.csv")
        """
        if header is None:
            header = ["Timestamp", "Skin Temperature", "Impedance", "Compensated Temperature"]
        self._csv_logger[Stream.TEMPERATURE] = CSVLogger(filename, header)

    def disable_csv_logging(self) -> None:
        """
        Stops logging stream data into CSV.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            x = application.disable_csv_logging()
        """
        if self._csv_logger.get(Stream.TEMPERATURE):
            self._csv_logger[Stream.TEMPERATURE].stop_logging()
        self._csv_logger[Stream.TEMPERATURE] = None

    def set_correction_factor(self, correction_factor: float):
        """
        Set correction factor.

        :return: None

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            application.set_correction_factor(2.5)

        """
        self._correction_factor = correction_factor

    def set_slope(self, slope: float):
        """
        Set slope.

        :return: None

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_temperature_application()
            application.set_slope(1.5)
        """
        self._slope = slope



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="adi_study_watch",
    version="4.1.0",
    author="Analog Devices, Inc.",
    author_email="healthcare-support@analog.com",
    license='Apache License, Version 2.0',
    description="ADI study watch python SDK",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/analogdevicesinc/study-watch-sdk",
    packages=["adi_study_watch",
              "adi_study_watch.core",
              "adi_study_watch.core.enums",
              "adi_study_watch.application",
              "adi_study_watch.core.packets",
              "adi_study_watch.core.data_types",
              "adi_study_watch.application.vsm_mb_sb",
              "adi_study_watch.application.study_watch",
              ],
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Intended Audience :: Healthcare Industry',
        'Topic :: Software Development :: Libraries',
    ],
    python_requires='>=3.6',
    install_requires=['pyserial==3.5', 'tqdm==4.61.0', 'pywinusb==0.4.2'],
)



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

import logging
from typing import List, Dict

from ..core import utils
from .csv_logging import CSVLogger
from .common_stream import CommonStream
from ..core.enums.dcb_enums import DCBCommand
from ..core.packets.command_packet import CommandPacket
from ..core.packets.stream_data_packets import BIADataPacket
from ..core.enums.common_enums import Application, Stream, CommonCommand
from ..core.enums.bia_enums import BIACommand, HSResistorTIA, BIADFTWindow
from ..core.packets.bia_packets import HSRTIAPacket, BIADCBPacket, BIADFTPacket, BIALibraryConfigPacket

logger = logging.getLogger(__name__)


class BIAApplication(CommonStream):
    """
    BIA Application class.

    .. code-block:: python3
        :emphasize-lines: 4

        from adi_study_watch import SDK

        sdk = SDK("COM4")
        application = sdk.get_bia_application()

    """

    RESISTOR_200 = HSResistorTIA.RESISTOR_200
    RESISTOR_1K = HSResistorTIA.RESISTOR_1K
    RESISTOR_5K = HSResistorTIA.RESISTOR_5K

    DFT_WINDOW_4 = BIADFTWindow.DFT_WINDOW_4
    DFT_WINDOW_8 = BIADFTWindow.DFT_WINDOW_8
    DFT_WINDOW_16 = BIADFTWindow.DFT_WINDOW_16
    DFT_WINDOW_32 = BIADFTWindow.DFT_WINDOW_32
    DFT_WINDOW_64 = BIADFTWindow.DFT_WINDOW_64
    DFT_WINDOW_128 = BIADFTWindow.DFT_WINDOW_128
    DFT_WINDOW_256 = BIADFTWindow.DFT_WINDOW_256
    DFT_WINDOW_512 = BIADFTWindow.DFT_WINDOW_512
    DFT_WINDOW_1024 = BIADFTWindow.DFT_WINDOW_1024
    DFT_WINDOW_2048 = BIADFTWindow.DFT_WINDOW_2048
    DFT_WINDOW_4096 = BIADFTWindow.DFT_WINDOW_4096
    DFT_WINDOW_8192 = BIADFTWindow.DFT_WINDOW_8192
    DFT_WINDOW_16384 = BIADFTWindow.DFT_WINDOW_16384

    def __init__(self, packet_manager):
        super().__init__(Application.BIA, Stream.BIA, packet_manager)
        self._dcb_size = 18

    @staticmethod
    def get_supported_dft_windows() -> List[BIADFTWindow]:
        """
        List all supported DFT window for BIA.

        :return: Array of DFT window enums.
        :rtype: List

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_bia_application()
            x = application.get_supported_dft_windows()
            print(x)
            # [<BIADFTWindow.DFT_WINDOW_4: ['0x0', '0x0']>, ... , <BIADFTWindow.DFT_WINDOW_16384: ['0x0', '0x12']>]
        """
        return [BIAApplication.DFT_WINDOW_4, BIAApplication.DFT_WINDOW_8, BIAApplication.DFT_WINDOW_16,
                BIAApplication.DFT_WINDOW_32, BIAApplication.DFT_WINDOW_64, BIAApplication.DFT_WINDOW_128,
                BIAApplication.DFT_WINDOW_256, BIAApplication.DFT_WINDOW_512, BIAApplication.DFT_WINDOW_1024,
                BIAApplication.DFT_WINDOW_2048, BIAApplication.DFT_WINDOW_4096, BIAApplication.DFT_WINDOW_8192,
                BIAApplication.DFT_WINDOW_16384]

    @staticmethod
    def get_supported_hs_resistor_tia_ids() -> List[HSResistorTIA]:
        """
        List all supported High Speed Resistor Trans Impedance Amplifier for BIA.

        :return: Array of High Speed Resistor Trans Impedance Amplifier enums.
        :rtype: List

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_bia_application()
            x = application.get_supported_hs_resistor_tia_ids()
            print(x)
            # [<HSResistorTIA.RESISTOR_200: ['0x0', '0x0']>, ... , <HSResistorTIA.RESISTOR_5K: ['0x0', '0x2']>]

        """
        return [BIAApplication.RESISTOR_200, BIAApplication.RESISTOR_1K, BIAApplication.RESISTOR_5K]

    def calibrate_hs_resistor_tia(self, hs_resistor_tia_id: HSResistorTIA) -> Dict:
        """
        Calibrate High Speed Resistor Trans Impedance Amplifier.

        :param hs_resistor_tia_id: High Speed Resistor Trans Impedance Amplifier to calibrate,
                                  | use get_supported_hs_resistor_tia_ids() to list all supported resistor.
        :type hs_resistor_tia_id: HSResistorTIA
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_bia_application()
            x = application.get_supported_hs_resistor_tia_ids()
            print(x)
            # [<HSResistorTIA.RESISTOR_200: ['0x0', '0x0']>, ... , <HSResistorTIA.RESISTOR_5K: ['0x0', '0x2']>]
            x = application.calibrate_hs_resistor_tia(application.RESISTOR_1K)
            print(x["payload"]["hs_resistor_tia"])
            # HSResistorTIA.RESISTOR_1K

        """
        request_packet = HSRTIAPacket(self._destination, BIACommand.SET_HS_TRANS_IMPEDANCE_AMPLIFIER_CAL_REQ)
        request_packet.set_payload("hs_resistor_tia", hs_resistor_tia_id)
        response_packet = HSRTIAPacket(self._destination, BIACommand.SET_HS_TRANS_IMPEDANCE_AMPLIFIER_CAL_RES)
        return self._send_packet(request_packet, response_packet)

    def set_discrete_fourier_transformation(self, dft_window: BIADFTWindow) -> Dict:
        """
        Set Discrete Fourier Transformation for BIA.

        :param dft_window: DFT window for Discrete Fourier Transformation, use get_supported_dft_windows()
                          | to list all supported DFT window.
        :type dft_window: BIADFTWindow
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_bia_application()
            x = application.get_supported_dft_windows()
            print(x)
            # [<BCMDFTWindow.DFT_WINDOW_4: ['0x0', '0x0']>, ... ,<BCMDFTWindow.DFT_WINDOW_16384: ['0x0', '0x12']>]
            x = application.set_discrete_fourier_transformation(application.DFT_WINDOW_16384)
            print(x["payload"]["dft_window"])
            # BCMDFTWindow.DFT_WINDOW_16384
        """
        request_packet = BIADFTPacket(self._destination, BIACommand.SET_DFT_NUM_REQ)
        request_packet.set_payload("dft_window", dft_window)
        response_packet = BIADFTPacket(self._destination, BIACommand.SET_DFT_NUM_RES)
        return self._send_packet(request_packet, response_packet)

    def read_library_configuration(self, fields: List[int]) -> Dict:
        """
        Reads library configuration from specified field values.

        :param fields: List of field values to read.
        :type fields: List[int]
        :return: A response packet as dictionary
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Fields Lower Limit
             - Fields Upper Limit
           * - 0x00
             - 0x12

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_bia_application()
            x = application.read_library_configuration([0x00])
            print(x["payload"]["data"])
            # [['0x0', '0x0']]

        """
        data = [[field, 0] for field in fields]
        request_packet = BIALibraryConfigPacket(self._destination, CommonCommand.READ_LCFG_REQ)
        request_packet.set_payload("size", len(data))
        request_packet.set_payload("data", data)
        response_packet = BIALibraryConfigPacket(self._destination, CommonCommand.READ_LCFG_RES)
        return self._send_packet(request_packet, response_packet)

    def write_library_configuration(self, fields_values: List[List[int]]) -> Dict:
        """
        Writes library configuration from List of fields and values.

        :param fields_values: List of fields and values to write.
        :type fields_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Fields Lower Limit
             - Fields Upper Limit
           * - 0x00
             - 0x12

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_bia_application()
            x = application.write_library_configuration([[0x00, 0x1]])
            print(x["payload"]["data"])
            # [['0x0', '0x1']]

        """
        request_packet = BIALibraryConfigPacket(self._destination, CommonCommand.WRITE_LCFG_REQ)
        request_packet.set_payload("size", len(fields_values))
        request_packet.set_payload("data", fields_values)
        response_packet = BIALibraryConfigPacket(self._destination, CommonCommand.WRITE_LCFG_RES)
        return self._send_packet(request_packet, response_packet)

    def write_dcb_to_lcfg(self) -> Dict:
        """
        Writes Device configuration block data to library configuration.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_bia_application()
            x = application.write_dcb_to_lcfg()
            print(x["payload"]["status"])
            # CommonStatus.OK
        """
        request_packet = CommandPacket(self._destination, CommonCommand.SET_LCFG_REQ)
        response_packet = CommandPacket(self._destination, CommonCommand.SET_LCFG_RES)
        return self._send_packet(request_packet, response_packet)

    def delete_device_configuration_block(self) -> Dict:
        """
        Deletes BIA Device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_bia_application()
            application.delete_device_configuration_block()
        """
        request_packet = CommandPacket(self._destination, DCBCommand.ERASE_CONFIG_REQ)
        response_packet = BIADCBPacket(self._destination, DCBCommand.ERASE_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def read_device_configuration_block(self) -> Dict:
        """
        Returns entire device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_bia_application()
            x = application.read_device_configuration_block()
            print(x["payload"]["data"])
            # []

        """
        request_packet = CommandPacket(self._destination, DCBCommand.READ_CONFIG_REQ)
        response_packet = BIADCBPacket(self._destination, DCBCommand.READ_CONFIG_RES)
        response_dict = self._send_packet(request_packet, response_packet)
        response_dict["payload"]["data"] = utils.add_index_to_array(response_dict["payload"]["data"], to_hex=True)
        return response_dict

    def write_device_configuration_block(self, addresses_values: List[List[int]]) -> Dict:
        """
        Writes the device configuration block values of specified addresses.
        This function takes a list of addresses and values to write, and returns a response packet as
        dictionary containing addresses and values.

        :param addresses_values: List of addresses and values to write.
        :type addresses_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x00
             - 0x12

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_bia_application()
            x = application.write_device_configuration_block([[0x0, 2], [0x1, 0x1]])
            print(x["payload"]["size"])
            # 2

        """
        dcb_array = [0] * self._dcb_size
        for address_value in addresses_values:
            address = address_value[0]
            value = address_value[1]
            dcb_array[address] = value
        request_packet = BIADCBPacket(self._destination, DCBCommand.WRITE_CONFIG_REQ)
        request_packet.set_payload("size", len(dcb_array))
        request_packet.set_payload("data", dcb_array)
        response_packet = BIADCBPacket(self._destination, DCBCommand.WRITE_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def write_device_configuration_block_from_file(self, filename: str) -> Dict:
        """
        Writes the device configuration block values of specified addresses from file.

        :param filename: dcb filename
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x00
             - 0x12

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_bia_application()
            application.write_device_configuration_block_from_file("bcm_dcb.dcfg")

        """
        result = self.device_configuration_file_to_list(filename)
        if result:
            return self.write_device_configuration_block(result)

    def _callback_data(self, packet, packet_id, callback_function=None, args=None):
        """
        Process and returns the data back to user's callback function.
        """
        self._callback_data_helper(packet, BIADataPacket())

    def enable_csv_logging(self, filename, header=None) -> None:
        """
        Start logging stream data into CSV.

        :param filename: Name of the CSV file.
        :param header: Header list of the CSV file.
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_bia_application()
            x = application.enable_csv_logging("bcm.csv")
        """
        if header is None:
            header = ["Timestamp", "IMP Real(Ohms)", "IMP Img(Ohms)", "IMP Module(Ohms)", "IMP Phase(Rad)",
                      "ADM Real(Ohms)", "ADM Img(Ohms)", "ADM Module(Ohms)", "ADM Phase(Rad)", "Seq No."]
        self._csv_logger[Stream.BIA] = CSVLogger(filename, header)

    def disable_csv_logging(self) -> None:
        """
        Stops logging stream data into CSV.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_bia_application()
            x = application.disable_csv_logging()
        """
        if self._csv_logger.get(Stream.BIA):
            self._csv_logger[Stream.BIA].stop_logging()
        self._csv_logger[Stream.BIA] = None



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from ..data_types.array import Array
from ..data_types.enums import Enums
from ..data_types.integer import Int
from ..data_types.binary import Binary
from .command_packet import CommandPacket
from ..enums.pm_enums import MCUType, ElectrodeSwitch, LDO, ChipID, BatteryStatus, PowerMode


class AppsHealthPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PM: ['0xC5', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x16',
                'checksum': '0x0'
            },
            'payload': {
                'command': <PMCommand.GET_APPS_HEALTH_RES: ['0x91']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'ad5940_isr_count': 202,
                'adpd_isr_count': 3350,
                'adxl_isr_count': 1187
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["ad5940_isr_count"] = Int(4)
        self._config["payload"]["adpd_isr_count"] = Int(4)
        self._config["payload"]["adxl_isr_count"] = Int(4)


class BatteryInfoPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PM: ['0xC5', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x12',
                'checksum': '0x0'
            },
            'payload': {
                'command': <PMCommand.GET_BAT_INFO_RES: ['0x45']>,
                'status': <PMStatus.OK: ['0x41']>,
                'timestamp': 668880834,
                'battery_status': <BatteryStatus.COMPLETE: ['0x03']>,
                'battery_level': 100,
                'battery_mv': 4334
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["timestamp"] = Int(4)
        self._config["payload"]["battery_status"] = Enums(1, enum_class=BatteryStatus)
        self._config["payload"]["battery_level"] = Int(1)
        self._config["payload"]["battery_mv"] = Int(2)


class BatteryThresholdPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PM: ['0xC5', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xA',
                'checksum': '0x0'
            },
            'payload': {
                'command': <PMCommand.SET_BAT_THR_RES: ['0x47']>,
                'status': <PMStatus.OK: ['0x41']>
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["low_level"] = Int(1, value_limit=[0, 100])
        self._config["payload"]["critical_level"] = Int(1, value_limit=[0, 100])


class ChipIDPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PM: ['0xC5', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xD',
                'checksum': '0x0'
            },
            'payload': {
                'command': <PMCommand.CHIP_ID_RES: ['0x7D']>,
                'status': <PMStatus.OK: ['0x41']>,
                'chip_name': <ChipID.ADPD4K: ['0x02']>,
                'chip_id': 192
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["chip_name"] = Enums(1, enum_class=ChipID)
        self._config["payload"]["chip_id"] = Int(2)


class SwitchControlPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PM: ['0xC5', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xC',
                'checksum': '0x0'
            },
            'payload': {
                'command': <PMCommand.SW_CONTROL_RES: ['0x79']>,
                'status': <PMStatus.OK: ['0x41']>,
                'switch_name': <ElectrodeSwitch.ADPD4000: ['0x02']>,
                'enabled': False
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["switch_name"] = Enums(1, enum_class=ElectrodeSwitch)
        self._config["payload"]["enabled"] = Binary(1)


class LDOControlPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PM: ['0xC5', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xC',
                'checksum': '0x0'
            },
            'payload': {
                'command': <PMCommand.LDO_CONTROL_RES: ['0x7B']>,
                'status': <PMStatus.OK: ['0x41']>,
                'ldo_name': <LDO.OPTICAL: ['0x02']>,
                'enabled': False
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["ldo_name"] = Enums(1, enum_class=LDO)
        self._config["payload"]["enabled"] = Binary(1)


class ControlPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PM: ['0xC5', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xB',
                'checksum': '0x0'
            },
            'payload': {
                'command': <PMCommand.CAP_SENSE_TEST_RES: ['0x7F']>,
                'status': <PMStatus.OK: ['0x41']>,
                'enabled': False
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["enabled"] = Binary(1)


class DateTimePacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PM: ['0xC5', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x15',
                'checksum': '0x0'
            },
            'payload': {
                'command': <PMCommand.GET_DATE_TIME_RES: ['0x53']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'year': 2021,
                'month': 3,
                'day': 12,
                'hour': 16,
                'minute': 33,
                'second': 12,
                'tz_sec': 19800
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["year"] = Int(2)
        self._config["payload"]["month"] = Int(1)
        self._config["payload"]["day"] = Int(1)
        self._config["payload"]["hour"] = Int(1)
        self._config["payload"]["minute"] = Int(1)
        self._config["payload"]["second"] = Int(1)
        self._config["payload"]["tz_sec"] = Int(4, sign=True)


class DCBStatusPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PM: ['0xC5', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x1A',
                'checksum': '0x0'
            },
            'payload': {
                'command': <DCBCommand.QUERY_STATUS_RES: ['0x9E']>,
                'status': <PMStatus.OK: ['0x41']>,
                'general_block': 0,
                'ad5940_block': 0,
                'adpd_block': 0,
                'adxl_block': 0,
                'ppg_block': 0,
                'ecg_block': 0,
                'eda_block': 0,
                'ad7156_block': 1,
                'pedometer_block': 0,
                'temperature_block': 0,
                'low_touch_block': 0,
                'ui_config_block': 0,
                'user0_block': 0,
                'user1_block': 0,
                'user2_block': 0,
                'user3_block': 0
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["general_block"] = Binary(1)
        self._config["payload"]["ad5940_block"] = Binary(1)
        self._config["payload"]["adpd_block"] = Binary(1)
        self._config["payload"]["adxl_block"] = Binary(1)
        self._config["payload"]["ppg_block"] = Binary(1)
        self._config["payload"]["ecg_block"] = Binary(1)
        self._config["payload"]["eda_block"] = Binary(1)
        self._config["payload"]["ad7156_block"] = Binary(1)
        self._config["payload"]["pedometer_block"] = Binary(1)
        self._config["payload"]["temperature_block"] = Binary(1)
        self._config["payload"]["low_touch_block"] = Binary(1)
        self._config["payload"]["ui_config_block"] = Binary(1)
        self._config["payload"]["user0_block"] = Binary(1)
        self._config["payload"]["user1_block"] = Binary(1)
        self._config["payload"]["user2_block"] = Binary(1)
        self._config["payload"]["user3_block"] = Binary(1)
        self._config["payload"]["bcm_block"] = Binary(1)


class MCUVersionPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PM: ['0xC5', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xB',
                'checksum': '0x0'
            },
            'payload': {
                'command': <PMCommand.GET_MCU_VERSION_RES: ['0x59']>,
                'status': <PMStatus.OK: ['0x41']>,
                'mcu': <MCUType.MCU_M4: ['0x02']>
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["mcu"] = Enums(1, enum_class=MCUType)


class PingPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PM: ['0xC5', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x49',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.PING_RES: ['0x1B']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'sequence_num': 1,
                'data': [ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, ... , 0, 0, 0, 0 ],
                'elapsed_time': 0.016231060028076172
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["sequence_num"] = Int(4)
        self._config["payload"]["data"] = Array(-1, data_types=[Int(1)])


class PowerStatePacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PM: ['0xC5', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xB',
                'checksum': '0x0'
            },
            'payload': {
                'command': <PMCommand.SET_POWER_STATE_RES: ['0x49']>,
                'status': <PMStatus.OK: ['0x41']>,
                'power_state': <PowerMode.ACTIVE: ['0x00']>
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["power_state"] = Enums(1, enum_class=PowerMode)


class SystemInfoPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PM: ['0xC5', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x24',
                'checksum': '0x0'
            },
            'payload': {
                'command': <PMCommand.SYS_INFO_RES: ['0x4B']>,
                'status': <PMStatus.OK: ['0x41']>,
                'version': 0,
                'mac_address': 'C5-05-CA-F1-67-D5',
                'device_id': 0,
                'model_number': 0,
                'hw_id': 0,
                'bom_id': 0,
                'batch_id': 0,
                'date': 0,
                'board_type': 0
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["version"] = Int(2)
        self._config["payload"]["mac_address"] = Array(6, data_types=[Int(1)])
        self._config["payload"]["device_id"] = Int(4)
        self._config["payload"]["model_number"] = Int(4)
        self._config["payload"]["hw_id"] = Int(2)
        self._config["payload"]["bom_id"] = Int(2)
        self._config["payload"]["batch_id"] = Int(1)
        self._config["payload"]["date"] = Int(4)
        self._config["payload"]["board_type"] = Int(1)

    def get_dict(self, last_timestamp=None):
        packet = super().get_dict(last_timestamp)
        packet["payload"]["mac_address"] = ''.join(
            '{:02X}-'.format(x) for x in packet["payload"]["mac_address"])[:-1]
        return packet


class EEPROMPacket(CommandPacket):
    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["size"] = Int(1)
        self._config["payload"]["data"] = Array(16, data_types=[Int(1, to_hex=True)])


class PMRegisterWritePacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PM: ['0xC5', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 15,
                'checksum': 0
            },
            'payload': {
                'command': <CommonCommand.REGISTER_WRITE_RES: ['0x24']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'size': 1,
                'data': [
                    [ '0x0', '0x10' ]
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["size"] = Int(1)
        self._config["payload"]["data"] = Array(-1, dimension=2,
                                                data_types=[
                                                    Int(2, to_hex=True),
                                                    Int(2, to_hex=True)
                                                ])


class PMRegisterReadPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PM: ['0xC5', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 15,
                'checksum': 0
            },
            'payload': {
                'command': <CommonCommand.REGISTER_READ_RES: ['0x22']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'size': 1,
                'data': [
                    [ '0x0', '0x10' ]
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["size"] = Int(1)
        self._config["payload"]["data"] = Array(-1, dimension=2,
                                                data_types=[
                                                    Int(2, value_limit=[0x00, 0x36], to_hex=True),
                                                    Int(2, to_hex=True)
                                                ])


class PMDCBPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PM: ['0xC5', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 32,
                'checksum': 0
            },
            'payload': {
                'command': <DCBCommand.READ_CONFIG_RES: ['0x98']>,
                'status': <DCBStatus.OK: ['0x97']>,
                'size': 5,
                'data': [
                    [ '0x0', '0x1B58' ],
                    [ '0x1', '0x1388' ],
                    [ '0x2', '0x564' ],
                    [ '0x3', '0x53C' ],
                    [ '0x4', '0x3' ]
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["size"] = Int(2)
        self._config["payload"]["data"] = Array(-1, dimension=2,
                                                data_types=[
                                                    Int(2, to_hex=True),
                                                    Int(2, to_hex=True)
                                                ], reverse_inner_array=True)



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************
import math
from typing import Dict, List

from ..core.enums.fs_enums import FSCommand
from ..core.enums.dcb_enums import DCBCommand
from .common_application import CommonApplication
from ..core.packets.command_packet import CommandPacket
from ..core.enums.low_touch_enum import LTCommand, CommandType
from ..core.enums.common_enums import Application, CommonCommand
from ..core.packets.low_touch_packets import ReadCH2CapPacket, CommandLogPacket, LTLibraryConfigPacket, LTDCBPacket


class LowTouchApplication(CommonApplication):
    """
    FS Application class.

    .. code-block:: python3
        :emphasize-lines: 4

        from adi_study_watch import SDK

        sdk = SDK("COM4")
        application = sdk.get_low_touch_application()
    """

    START_COMMAND = CommandType.START
    STOP_COMMAND = CommandType.STOP

    def __init__(self, packet_manager):
        super().__init__(Application.LT_APP, packet_manager)
        self._dcb_size = 57
        self._log_command = None

    def read_ch2_cap(self) -> Dict:
        """
        Read the AD7156 CH2 Capacitance in uF.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_low_touch_application()
            x = application.read_ch2_cap()
            print(x["payload"]["cap_value"])
            # 0
        """
        request_packet = ReadCH2CapPacket(self._destination, LTCommand.READ_CH2_CAP_REQ)
        response_packet = ReadCH2CapPacket(self._destination, LTCommand.READ_CH2_CAP_RES)
        return self._send_packet(request_packet, response_packet)

    def write_library_configuration(self, fields_values: List[List[int]]) -> Dict:
        """
        Writes library configuration from List of fields and values.

        :param fields_values: List of fields and values to write.
        :type fields_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Fields Lower Limit
             - Fields Upper Limit
           * - 0x00
             - 0x04

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_low_touch_application()
            x = application.write_library_configuration([[0x00, 0x1]])
            print(x["payload"]["data"])
            # [['0x0', '0x1']]

        """
        request_packet = LTLibraryConfigPacket(self._destination, CommonCommand.WRITE_LCFG_REQ)
        request_packet.set_payload("size", len(fields_values))
        request_packet.set_payload("data", fields_values)
        response_packet = LTLibraryConfigPacket(self._destination, CommonCommand.WRITE_LCFG_RES)
        return self._send_packet(request_packet, response_packet)

    def read_library_configuration(self, fields: List[int]) -> Dict:
        """
        Reads library configuration from specified field values.

        :param fields: List of field values to read.
        :type fields: List[int]
        :return: A response packet as dictionary
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Fields Lower Limit
             - Fields Upper Limit
           * - 0x00
             - 0x04

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_low_touch_application()
            x = application.read_library_configuration([0x00])
            print(x["payload"]["data"])
            # [['0x0', '0x0']]
        """
        data = [[field, 0] for field in fields]
        request_packet = LTLibraryConfigPacket(self._destination, CommonCommand.READ_LCFG_REQ)
        request_packet.set_payload("size", len(data))
        request_packet.set_payload("data", data)
        response_packet = LTLibraryConfigPacket(self._destination, CommonCommand.READ_LCFG_RES)
        return self._send_packet(request_packet, response_packet)

    def read_device_configuration_block(self, readable_format=False) -> [Dict]:
        """
        Returns entire device configuration block.

        :param readable_format: Converts binary result into readable commands.
        :return: A response packet as dictionary.
        :rtype: [Dict]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_low_touch_application()
            x = application.read_device_configuration_block()
            print(x["payload"]["data"])
        """
        request_packet = CommandPacket(self._destination, DCBCommand.READ_CONFIG_REQ)
        response_packet = LTDCBPacket(self._destination, DCBCommand.READ_CONFIG_RES)
        result = self._send_packet_multi_response(request_packet, response_packet)

        if readable_format:
            raw_data = []
            total_size = 0
            for packet in result:
                total_size += packet["payload"]["size"]
                raw_data += packet["payload"]["data"][:packet["payload"]["size"] * 4]

            dcb_readable_format = []
            data_read = 0
            start_cmd_count = 0
            stop_cmd_count = 0
            if len(raw_data) > 0:
                raw_data[0:2], raw_data[2:4] = raw_data[2:4], raw_data[0:2]
                log_packet = CommandLogPacket()
                log_packet.decode_packet(raw_data)
                log_packet = log_packet.get_dict()
                raw_data = log_packet["payload"]["commands"]
                start_cmd_count = log_packet["payload"]["start_cmd_count"]
                stop_cmd_count = log_packet["payload"]["stop_cmd_count"]

                while data_read < len(raw_data):
                    pkt = raw_data[data_read:data_read + 8]
                    if not len(pkt) == 8:
                        break
                    length = (int(pkt[4]) << 8) + int(pkt[5])
                    pkt = raw_data[data_read:data_read + length]
                    pkt[0:2], pkt[2:4] = pkt[2:4], pkt[0:2]
                    packet = CommandPacket()
                    packet.decode_packet(pkt)
                    packet = packet.get_dict()
                    dcb_readable_format.append({
                        "application": packet["header"]["source"],
                        "command": packet["payload"]["command"]
                    })
                    data_read += length
            result = result[0]
            result["payload"]["size"] = total_size
            result["payload"]["data"] = dcb_readable_format
            result["payload"]["start_command_count"] = start_cmd_count
            result["payload"]["stop_command_count"] = stop_cmd_count
        return result

    def _write_device_configuration_block(self, commands) -> [Dict]:
        dcb_size = (self._dcb_size * 4)
        packets = math.ceil(len(commands) / dcb_size)
        if packets > 18:
            raise Exception("Can't write more than 18 packets. Size limit 4104 bytes.")
        result = []
        for packet in range(packets):
            temp_command = commands[packet * dcb_size:(packet + 1) * dcb_size]
            request_packet = LTDCBPacket(self._destination, DCBCommand.WRITE_CONFIG_REQ)
            request_packet.set_payload("size", len(temp_command) // 4)
            request_packet.set_payload("packet_count", packets)
            request_packet.set_payload("data", temp_command)
            response_packet = LTDCBPacket(self._destination, DCBCommand.WRITE_CONFIG_RES)
            result.append(self._send_packet(request_packet, response_packet))
        return result

    def write_device_configuration_block_from_file(self, filename: str) -> [Dict]:
        """
        Writes the device configuration block values from specified binary file.

        :param filename: binary filename
        :return: A response packet as dictionary.
        :rtype: [Dict]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_low_touch_application()
            application.write_device_configuration_block_from_file("lt_dcb.dcb")
        """
        with open(filename, 'rb') as file:
            data = file.readlines()
            result = []
            for value in data:
                result += value
            return self._write_device_configuration_block(result)

    def delete_device_configuration_block(self) -> Dict:
        """
        Deletes ADPD Device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_low_touch_application()
            application.delete_device_configuration_block()
        """
        request_packet = CommandPacket(self._destination, DCBCommand.ERASE_CONFIG_REQ)
        response_packet = LTDCBPacket(self._destination, DCBCommand.ERASE_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def enable_command_logging(self, command_type: CommandType):
        """
        Starts recording SDK commands to a file.

        :param command_type: Start or Stop command recording.
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_low_touch_application()
            application.enable_command_logging(application.START)
        """
        if command_type == CommandType.START:
            self._log_command = CommandLogPacket(Application.FS, command=FSCommand.LOG_USER_CONFIG_DATA_REQ)
            self._packet_manager.subscribe_command_logger(self._log_command)
        self._packet_manager.enable_command_logging(command_type)

    def disable_command_logging(self, command_type: CommandType, filename="commands.LOG"):
        """
        Stops recording SDK commands to a file.

        :param command_type: Start or Stop command recording.
        :param filename: Name of the file to store commands.
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_low_touch_application()
            application.disable_command_logging(application.START)
        """
        self._packet_manager.disable_command_logging(command_type)
        if command_type == CommandType.STOP:
            self._packet_manager.unsubscribe_command_logger()
            self._generate_log_file(filename)

    def _generate_log_file(self, filename):
        self._log_command.set_header("source", self._packet_manager.source)
        with open(filename, 'wb') as file:
            commands = bytearray(self._log_command.to_list())
            remaining = len(commands) % 4
            if remaining:
                commands += b'\0' * (4 - remaining)
            file.write(commands)



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

import re
import logging
from queue import Queue

from ..core import utils

logger = logging.getLogger(__name__)


class CommonApplication:
    """
    A Common Application class.
    """

    def __init__(self, destination, packet_manager):
        """
        Initialize the common application class.

        :param destination: Address of the application.
        :param packet_manager: PacketManager Object.
        """
        self._timeout = 10
        self._packet_queues = {}
        self._destination = destination
        self._packet_manager = packet_manager

    def _get_packet_id(self, response_command, destination=None):
        """
        Generates a unique packet ID.
        """
        destination = destination if destination else self._destination
        packet_id = utils.join_multi_length_packets(destination.value + response_command.value)
        return packet_id

    def _get_queue(self, packet_id):
        """
        Returns specified queue for packet ID.
        """
        queue = self._packet_queues.get(packet_id, None)
        if not queue:
            self._packet_queues[packet_id] = Queue()
        return self._packet_queues[packet_id]

    def _send_packet(self, request_packet, response_packet=None, callback=None, get_bytes=False):
        """
        Sends the specified packet to the packet manager, subscribe the callback for the packet id,
        retrieves the packet response from queue and unsubscribe from the packet id.
        """
        callback = callback if callback else self._callback_command
        command = response_packet.get_payload("command")
        destination = response_packet.get_header("destination")
        packet_id = self._get_packet_id(command, destination)
        self._packet_manager.subscribe(packet_id, callback)
        queue = self._get_queue(packet_id)
        self._packet_manager.send_packet(request_packet)
        response_byte = self._get_queue_data(queue)
        response_packet.decode_packet(response_byte)
        self._packet_manager.unsubscribe(packet_id, callback)
        response_dict = response_packet.get_dict()
        logger.debug(f"Packet decoded : {response_dict}")
        if get_bytes:
            return response_dict, response_byte
        return response_dict

    def _send_packet_no_response(self, request_packet, ):
        """
        Sends the specified packet to the packet manager, subscribe the callback for the packet id,
        retrieves the packet response from queue and unsubscribe from the packet id.
        """
        self._packet_manager.send_packet(request_packet)
        return {}

    def _send_packet_multi_response(self, request_packet, response_packet=None, packet_limit=None):
        result = []
        packet_count = 1
        command = response_packet.get_payload("command")
        destination = response_packet.get_header("destination")
        packet_id = self._get_packet_id(command, destination)
        self._packet_manager.subscribe(packet_id, self._callback_command)
        queue = self._get_queue(packet_id)
        self._packet_manager.send_packet(request_packet)
        while True:
            data = self._get_queue_data(queue)
            temp_packet = response_packet.__class__()
            temp_packet.decode_packet(data)
            packet_dict = temp_packet.get_dict()
            result.append(packet_dict)
            number_of_packets = packet_dict["payload"]["packet_count"]
            if number_of_packets == packet_count or number_of_packets == 0:
                break
            if packet_limit and packet_limit <= packet_count:
                break
            packet_count += 1
        self._packet_manager.unsubscribe(packet_id, self._callback_command)
        return result

    def _get_queue_data(self, queue):
        response_byte = [0, 0, 0, 0, 0, 10, 0, 0, -1, -1] + [0] * 256
        try:
            response_byte = queue.get(timeout=self._timeout)
        except Exception as e:
            logger.debug(f"No data received, reason :: {e}.")
        return response_byte

    def _callback_command(self, data, packet_id):
        """
        Receives the command response packet and store it in its respective packet id queue.
        """
        self._get_queue(packet_id).put(data)

    def set_timeout(self, timeout_value: float):
        """
        Sets the time out for queue to wait for command packet response.

        :param timeout_value: queue timeout value.
        :type timeout_value: int
        """
        self._timeout = timeout_value

    @staticmethod
    def _parse_single_dcb_line(dcb_line):
        str_val = dcb_line.split('#')
        str_val[0] = re.sub(r'[ ]+', " ", str_val[0], 0, re.MULTILINE)
        dcb = str_val[0].replace('\t', '').replace('\n', '').split(" ")
        return dcb

    @staticmethod
    def device_configuration_file_to_list(dcfg_file, address=True):
        try:
            file = open(dcfg_file, 'r')
            result = []
            for line in file:
                if line[0] != '#' and line[0] != '\n' and line[0] != ' ' and line[0] != '\t':
                    dcb = CommonApplication._parse_single_dcb_line(line)
                    if address:
                        result.append([int(dcb[0], 16), int(dcb[1], 16)])
                    else:
                        result.append(int(dcb[0], 16))
            file.close()
            return result
        except Exception as e:
            logger.error(f"Can't parse {dcfg_file} file, make sure it's a single app dcfg file. Error :: {e}.",
                         exc_info=True)



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************
import logging

from .. import utils
from ..data_types.enums import Enums
from ..data_types.integer import Int
from ..enums import common_enums
from ..enums.common_enums import Application, CommonCommand, CommonStatus, Stream

logger = logging.getLogger(__name__)


class CommandPacket:
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.BIA: ['0xC3', '0x07']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xC',
                'checksum': '0x0'
            },
            'payload': {
                'command': <BIACommand.SET_HS_TRANS_IMPEDANCE_AMPLIFIER_CAL_RES: ['0x49']>,
                'status': <CommonStatus.OK: ['0x00']>,
            }
        }
    """

    def __init__(self, destination=None, command=None):
        self._config = dict(header={}, payload={})
        self._config["header"]["source"] = Enums(2, Application)
        self._config["header"]["destination"] = Enums(2, Application, default=destination)
        self._config["header"]["length"] = Int(2, reverse=True)
        self._config["header"]["checksum"] = Int(2, reverse=True)
        self._config["payload"]["command"] = Enums(1, CommonCommand, default=command)
        self._config["payload"]["status"] = Enums(1, CommonStatus, default=CommonStatus.OK)

    def __str__(self):
        return str(self.get_dict())

    def set_header(self, key, value):
        self._config["header"][key].set_value(value)

    def set_payload(self, key, value):
        self._config["payload"][key].set_value(value)

    def get_header(self, key):
        return self._config["header"][key].get_value()

    def get_payload(self, key):
        return self._config["payload"][key].get_value()

    @staticmethod
    def decode(data, config, start_index):
        if config.get_size() == -1:
            data_slice = data[start_index:]
        else:
            data_slice = data[start_index: start_index + config.get_size()]
        config.decode(data_slice)

    def decode_packet(self, data):
        command = data[8:9]
        source = data[0:2]
        # change source config for stream.
        if command == CommonCommand.STREAM_DATA.value or source == Stream.FS.value:
            self._config["header"]["source"] = Enums(2, Stream)

        start_index = 0
        for key in self._config["header"]:
            self.decode(data, self._config["header"][key], start_index)
            start_index += self._config["header"][key].get_size()

        for key in self._config["payload"]:
            if key == "command":
                # change in command config.
                source = self.get_header("source")
                self._config["payload"]["command"] = Enums(1, common_enums.get_command(command, source))
            elif key == "status":
                # change in status config.
                status = data[9:10]
                source = self.get_header("source")
                self._config["payload"]["status"] = Enums(1, common_enums.get_status(status, source, command))
            self.decode(data, self._config["payload"][key], start_index)
            start_index += self._config["payload"][key].get_size()
        # if not len(data) == start_index:
        #     print(data)
        #     logger.warning(f"There is still some data left to parse. {self.get_dict()}")

    def get_id(self):
        subscribe_id = []
        subscribe_id += self._config["header"]["source"].encode()
        subscribe_id += self._config["payload"]["command"].encode()
        return utils.join_multi_length_packets(subscribe_id)

    def _generate_dict(self):
        packet = dict(header={}, payload={})
        for key in self._config["header"]:
            packet["header"][key] = self._config["header"][key].get_value()
        for key in self._config["payload"]:
            packet["payload"][key] = self._config["payload"][key].get_value()
        return packet

    @staticmethod
    def _trim_data_array(packet):
        size_present = packet["payload"].get("size", None)
        data_present = packet["payload"].get("data", None)
        if size_present is not None and data_present is not None:
            if packet["payload"]["size"] == 0:
                packet["payload"]["data"] = []
            else:
                packet["payload"]["data"] = packet["payload"]["data"][:packet["payload"]["size"]]
        return packet

    def get_dict(self, last_timestamp=None):
        packet = self._generate_dict()
        return self._trim_data_array(packet)

    def to_list(self):
        packet_list = []
        for key in self._config["header"]:
            packet_list += self._config["header"][key].encode()
        for key in self._config["payload"]:
            packet_list += self._config["payload"][key].encode()
        self.set_header("length", len(packet_list))
        packet_list[4:6] = self._config["header"]["length"].encode()
        return packet_list
